/**
 * For copyright see license.txt.
 */
package anasy.examples.kalq2;

import java.util.HashMap;
import java.util.Map;

import anasy.Node;


@SuppressWarnings("serial")
class SemanticError extends Exception {
	public SemanticError(String msg) {
		super(msg);
	}
}


class MyNode implements anasy.Node {
	public MyNode eval() throws SemanticError {
		return this;
	}

	public MyNode doUnaryOp(String op) throws SemanticError {
		throw new SemanticError(String.format("Operation '%s' not implemented.", op));
	}

	public MyNode doBinOp(String op, MyNode other) throws SemanticError {
		if (";".equals(op)) {
			eval();
			return other.eval();
		}
		throw new SemanticError(String.format("Operation '%s' not implemented.", op));
	}

	public boolean isTrue() {
		return true;
	}
}


class Num extends MyNode {
	double val;

	public Num(double val) {
		super();
		this.val = val;
	}

	@Override
	public String toString() {
		return Double.toString(val);
	}

	@Override
	public MyNode doUnaryOp(String op) throws SemanticError {
		if ("-".equals(op))
			return new Num(-this.val);
		if ("+".equals(op))
			return this;
		if ("--".equals(op))
			return new Num(this.val - 1);
		if ("++".equals(op))
			return new Num(this.val + 1);
		if ("!".equals(op)) {
			double n = val, r = 1;
			while (n > 1) r *= n--;
			return new Num(r);
		}
		if ("@".equals(op))
			return new Num(this.hashCode());
		return super.doUnaryOp(op);
	}

	@Override
	public MyNode doBinOp(String op, MyNode other) throws SemanticError {
		Num that = (Num)other;
		if ("+".equals(op))
			return new Num(val + that.val);
		if ("-".equals(op))
			return new Num(val - that.val);
		if ("*".equals(op))
			return new Num(val * that.val);
		if ("/".equals(op))
			return new Num(val / that.val);
		if ("%".equals(op))
			return new Num(val % that.val);
		if ("^".equals(op))
			return new Num(Math.pow(val, that.val));
		if ("==".equals(op))
			return new Num(val == that.val ? 1 : 0);
		if ("!=".equals(op))
			return new Num(val != that.val ? 1 : 0);
		if ("<=".equals(op))
			return new Num(val <= that.val ? 1 : 0);
		if ("<".equals(op))
			return new Num(val < that.val ? 1 : 0);
		if (">=".equals(op))
			return new Num(val >= that.val ? 1 : 0);
		if (">".equals(op))
			return new Num(val > that.val ? 1 : 0);
		return super.doBinOp(op, other);
	}

	@Override
	public boolean isTrue() {
		return this.val != 0;
	}
}


class UnaryOp extends MyNode {
	String op;
	MyNode operand;

	public UnaryOp(String op, Node operand) {
		super();
		this.op = op;
		this.operand = (MyNode) operand;
	}
	
	@Override
	public String toString() {
		return  "(" + op + " " + operand.toString() + ")";
	}
	
	@Override
	public MyNode eval() throws SemanticError {
		return operand.eval().doUnaryOp(op);
	}
}


class BinOp extends MyNode {
	String op;
	MyNode left;
	MyNode right;

	public BinOp(String op, Node left, Node right) {
		super();
		this.op = op;
		this.left = (MyNode) left;
		this.right = (MyNode) right;
	}

	@Override
	public String toString() {
		return "(" + left.toString() + op + right.toString() + ")";
	}
	
	@Override
	public MyNode eval() throws SemanticError {
		MyNode l = ("=".equals(op)) ? left : left.eval();
		MyNode r = right.eval();
		return l.doBinOp(op, r);
	}
}


class Branch extends MyNode {
	// if head then body else tail
	MyNode head;
	MyNode body;
	MyNode tail;
	
	public Branch(Node head, Node body, Node tail) {
		super();
		this.head = (MyNode) head;
		this.body = (MyNode) body;
		this.tail = (MyNode) tail;
	}
	
	@Override
	public String toString() {
		return "if " + head.toString() + " then " + body.toString() + " else "+ tail.toString() + " end";
	}

	@Override
	public MyNode eval() throws SemanticError {
		if (head.eval().isTrue())
			return body.eval();
		else
			return tail.eval();
	}
}


public class Engine {

	public static String itemsToString(MyNode[] items, String delim) {
		StringBuffer s = new StringBuffer();
		int i = 0;
		while (i < items.length) {
			s = s.append(items[i].toString());
			if (i < items.length - 1)
				s = s.append(delim);
			i++;
		}
		return s.toString();
	}

	class Context {
		Context parent;
		Map<Sym, MyNode> vars;

		public Context(Context parent) {
			this.parent = parent;
			this.vars = new HashMap<Sym, MyNode>();
		}
		
		public MyNode get(Sym sym) {
			if (vars.containsKey(sym))
				return vars.get(sym);
			if (parent != null)
				return parent.get(sym);
			return null;
		}

		public MyNode set(Sym sym, MyNode val) {
			vars.put(sym, val);
			return val;
		}
	}
	
	class Sym extends MyNode {
		String val;
	
		public Sym(String val) {
			super();
			this.val = val;
		}
	
		@Override
		public String toString() {
			return val;
		}
	
		@Override
		public MyNode eval() throws SemanticError {
			MyNode val = context.get(this);
			if (val == null)
				throw new SemanticError(String.format("Undefined variable '%s'", this));
			return val;
		}
		@Override
		public MyNode doBinOp(String op, MyNode other) throws SemanticError {
			if ("=".equals(op))
				return context.set(this, other);
			return super.doBinOp(op, other);
		}
	}

	class Composite extends MyNode {
		Sym head;
		MyNode[] args;

		public Composite(Sym head, Node[] args) {
			this.head = head;
			this.args = new MyNode[args.length];
			for (int i = 0; i < args.length; i++)
				this.args[i] = (MyNode) args[i];
		}
	
		@Override
		public String toString() {
			return this.head + "(" + itemsToString(args, ",") + ")";
		}

		@Override
		public MyNode eval() throws SemanticError {
			MyNode node = context.get(this.head);
			if (!(node instanceof Func))
				throw new SemanticError(String.format("'%s' is not a function.", this.head));
			Func func = (Func) node;
			if (func.args.length != this.args.length)
				throw new SemanticError("Argument mismatch.");
			push();
			for (int i = 0; i < func.args.length; i++)
				context.set( ((Sym)func.args[i]), args[i].eval());
			MyNode result;
			try {
				result = func.body.eval();
			} finally {
				pop();
			}
			return result;
		}	
	}

	class Func extends MyNode {
		protected Sym head;
		protected MyNode[] args;
		protected MyNode body;

		public Func(Sym head, Node[] args, Node body) {
			this.head = head;
			this.args = new MyNode[args.length];
			for (int i = 0; i < args.length; i++)
				this.args[i] = (MyNode) args[i];
			this.body = (MyNode) body;
		}

		@Override
		public MyNode eval() throws SemanticError {
			return context.set(head, this);
		}

		@Override
		public String toString() {
			return "fun " + head + "(" + itemsToString(args, ",") + ") is " + body + " end";
		}
	}

	// ***** Symbols *****

	public Num num(double d) {
		return new Num(d);
	}

	private Map<String, Sym> symbols;

	public Sym sym(String str) {
		if (symbols.containsKey(str))
			return symbols.get(str);
		Sym sym = new Sym(str);
		symbols.put(str, sym);
		return sym;
	}
	
	public BinOp binop(String name, Node left, Node right) {
		return new BinOp(name, left, right);
	}
	
	public UnaryOp unop(String name, Node operand) {
		return new UnaryOp(name, operand);
	}

	public Branch branch(Node head, Node body, Node tail) {
		return new Branch(head, body, tail);
	}

	public Func func(Sym head, Node[] args, Node body) {
		return new Func(head, args, body);
	}

	public Composite composite(Sym head, Node[] args) {
		return new Composite(head, args);
	}

	// Context
	Context context;
	
	public void push() {
		this.context = new Context(this.context);
	}

	public void pop() {
		this.context = this.context.parent;
	}

	// **** Constructor *****

	public Engine() {
		symbols = new HashMap<String, Sym>();
		context = new Context(null);
	}
}
