/**
 * For copyright see license.txt.
 */
package anasy.examples.kalq2;

/*
Start :=
	<begin> Main <end>
Main :=
	Expr
ExprIndented :=
	<indent> Expr <dedent>
Expr :=
	Num | Name
	'(' ExprIndented ')'
	Expr '==' Expr | Expr '!=' Expr | Expr '<' Expr | ...
	Expr '+' Expr | Expr '-' Expr | ...
	'+' Expr | '-' Expr
	
	ignore(WS)
	ignore(inline comment)
	'if' Term 'then' Term 'else' Term
	'if' Term 'then' ITerm 'else' ITerm
 
*/

import java.util.ArrayList;
import java.util.List;

import anasy.Node;
import anasy.SyntaxError;
import anasy.Token;
import anasy.Tokind;
import anasy.examples.kalq2.Engine.Composite;
import anasy.examples.kalq2.Engine.Sym;
import anasy.input.Lexer;
import anasy.input.Location;
import anasy.syntax.literals.Name;
import anasy.syntax.literals.RealNum;
import anasy.syntax.operators.Infix;
import anasy.syntax.operators.Operator;
import anasy.syntax.operators.Outfix;
import anasy.syntax.operators.Postfix;
import anasy.syntax.operators.Prefix;
import anasy.syntax.operators.PrefixInfix;
import anasy.syntax.operators.PrefixPostfix;
import anasy.syntax.special.Expression;
import anasy.syntax.whitespace.InlineComment;
import anasy.syntax.whitespace.Whitespace;


public class Parser2 {

	protected final Engine engine = new Engine();
	protected boolean canIndent = false;

	class TokInfix extends Infix {
		public TokInfix(String name, int lbp, boolean rightAssoc) {
			super(name, lbp, rightAssoc);
		}

		public TokInfix(String name, int lbp) {
			this(name, lbp, false);
		}

		@Override
		protected Node makeInfixNode(Node left, Node right) {
			return engine.binop(name, left, right);
		}		
	}

	class TokPrefixInfix extends PrefixInfix {
		public TokPrefixInfix(String name, int pbp, int lbp, boolean rightAssoc) {
			super(name, pbp, lbp, rightAssoc);
		}

		public TokPrefixInfix(String name, int pbp, int lbp) {
			this(name, pbp, lbp, false);
		}

		@Override
		public Node makeInfixNode(Node left, Node right) {
			return engine.binop(name, left, right);
		}

		@Override
		public Node makePrefixNode(Node operand) {
			return engine.unop(name, operand);
		}
	}
	
	class TokPrefixPostfix extends PrefixPostfix {
		public TokPrefixPostfix(String name, int pbp, int lbp) {
			super(name, pbp, lbp);
		}

		@Override
		public Node makePrefixNode(Node operand) {
			return engine.unop(name, operand);
		}

		@Override
		public Node makePostfixNode(Node operand) {
			return engine.unop(name, operand);
		}
	}

	class TokParentheses extends Outfix {
		public final Tokind tokComma = new Operator(",", 0);
		
		public TokParentheses() {
			super("(", ")", 200);
		}

		@Override
		public void setParent(Tokind parent) {
			super.setParent(parent);
			parent.add(tokComma);
		}
	
		@Override
		public Token parse(Token left, int rbp) throws SyntaxError {
			// infix form represents function call
			final Lexer lexer = getLexer();
			final Location loc = lexer.prev().loc;
			if (!(left.value instanceof Engine.Sym))
				error(loc, "Function name expected");
			List<Node> exprs = new ArrayList<Node>();
			if (!lexer.advanceIf(tokClose))
				do
					exprs.add(parent.parse().value);
				while (lexer.advanceIf(tokComma));
			lexer.advance(tokClose);
			Node[] nodes = exprs.toArray(new Node[exprs.size()]);
			return new Token(this, loc, engine.new Composite((Engine.Sym)left.value, nodes));
		}	
	}

	public class TokIf extends Operator {
		public final Tokind tokThen = new Operator("then", 0);
		public final Tokind tokElse = new Operator("else", 0);;

		public TokIf(int lbp) {
			super("if", lbp);
		}

		@Override
		public void setParent(Tokind parent) {
			super.setParent(parent);
			parent.add(tokThen, tokElse);
		}
		
		@Override
		public Token lex() throws SyntaxError {
			Token t = super.lex();
			if (t == null) t = tokThen.lex();
			if (t == null) t = tokElse.lex();
			return t;
		}

		public Node makeIfNode(Node head, Node body, Node tail) {
			return engine.branch(head, body, tail);
		}

		@Override
		public Token parse(int rbp) throws SyntaxError {
			// prefix form: if Head then Body else Tail end
			final Lexer lexer = getLexer();
			final Location loc = lexer.prev().loc;
			Node head = parent.parse(lbp).value;
			lexer.advance(tokThen);
			canIndent = true;
			Node body = parent.parse(tokThen.lbp).value;
			lexer.advance(tokElse);
			canIndent = true;
			Node tail = parent.parse(tokElse.lbp).value;
			return new Token(this, loc, makeIfNode(head, body, tail));
		}

		@Override
		public Token parse(Token left, int rbp) throws SyntaxError {
			// infix form: Body if Head else Tail
			final Lexer lexer = getLexer();
			final Location loc = lexer.prev().loc;
			Token head = parent.parse(lbp);
			lexer.advance(tokElse);
			Node tail = parent.parse(tokElse.lbp).value;
			return new Token(this, loc, makeIfNode(head.value, left.value, tail));
		}
	}

	class TokFun extends Operator {
		private final Tokind tokIs = new Operator("is", 0);

		@Override
		public Token lex() throws SyntaxError {
			Token t = super.lex();
			if (t == null) t = tokIs.lex();
			return t;
		}

		public TokFun(int lbp) {
			super("fun", lbp);
		}

		@Override
		public void setParent(Tokind parent) {
			super.setParent(parent);
			parent.add(tokIs);
		}

		@Override
		public Token parse(int rbp) throws SyntaxError {
			// fun <name> (<args>) is <body> end
			final Lexer lexer = getLexer();
			final Location loc = lexer.prev().loc;
			Token tok = parent.parse(0);
			if (!(tok.value instanceof Composite))
				error(tok.loc, "Arguments expected");
			lexer.advance(tokIs);
			Node body = parent.parse(0).value;
			Composite c = (Composite) tok.value;
			return new Token(this, loc, engine.func(c.head, c.args, body));
		}			
	}

	protected Tokind expr() {
		return new Expression("<expr>",
			new Whitespace(),
			new InlineComment(";;"),
			new RealNum() {
				@Override
				public Node makeNode(String lexeme) {
					return engine.num(Double.parseDouble(lexeme));
				}
			},
			new Name() {
				@Override
				public Node makeNode(String lexeme) {
					return engine.sym(lexeme);
				}
			},
			new TokParentheses(),
			new TokInfix(";", 10),
			new TokFun(20),
			new TokIf(30),
			new TokInfix("==", 40),
			new TokInfix("=", 20),
			new TokInfix("!=", 40),
			new TokInfix("<=", 40),
			new TokInfix("<", 40),
			new TokInfix(">=", 40),
			new TokInfix(">", 40),
			new TokPrefixInfix("+", 100, 50),
			new TokPrefixInfix("-", 100, 50),
	 		new TokInfix("*", 60),
			new TokInfix("/", 60),
			new TokInfix("%", 60),
			new TokInfix("^", 70, true),
			new TokPrefixPostfix("++", 80, 80),
			new TokPrefixPostfix("--", 80, 80),
			new Postfix("!", 80) {
				@Override
				public Node makePostfixNode(Node operand) {
					return engine.unop(name, operand);
				}
			},
			new Prefix("@", 80) {
				@Override
				public Node makePrefixNode(Node operand) {
					return engine.unop(name, operand);
				}			
			},
			new Operator("prefix", 0) {
				@Override
				public Token parse(int rbp) throws SyntaxError {
					// prefix name pbp
					Token tok = parent.parse(0);
					if (!(tok.value instanceof Sym)) error(tok.loc, "Symbol expected");
					final Sym sym = (Sym) tok.value;
					//
					tok = parent.parse(0);
					if (!(tok.value instanceof Num)) error(tok.loc, "Number expected");
					Num pbp = (Num) tok.value;
					//
					Prefix p = new Prefix(sym.val, (int)pbp.val) {
						@Override
						public Node makePrefixNode(Node operand) {
							Node[] args = { operand };
							return engine.composite(sym, args);
						}
					};
					parent.add(p);
					return new Token(this, tok.loc, engine.num(0));
				}
			},
			new Operator("infix", 0) {
				@Override
				public Token parse(int rbp) throws SyntaxError {
					// infix name lbp
					Token tok = parent.parse(0);
					if (!(tok.value instanceof Sym)) error(tok.loc, "Symbol expected");
					final Sym sym = (Sym) tok.value;
					//
					tok = parent.parse(0);
					if (!(tok.value instanceof Num)) error(tok.loc, "Number expected");
					Num lbp = (Num) tok.value;
					//
					Infix i = new Infix(sym.val, (int)lbp.val) {
						@Override
						public Node makeInfixNode(Node left, Node right) {
							Node[] args = { left, right };
							return engine.new Composite(sym, args);
						}
					};
					parent.add(i);
					return new Token(this, tok.loc, engine.num(0));
				}
			}
		);

	}

}

/*
@Override
public Token parse(int rbp) throws SyntaxError {
	final Lexer lexer = getLexer();
	if (canIndent) ensure();
	boolean indented = canIndent && lexer.advanceIf(indenter.tokIndent);
	canIndent = false;
	if (indented) rbp = 0;
	Token token = super.parse(rbp);
	if (indented) lexer.advance(indenter.tokDedent);
	return token;
}
*/