package anasy;

import anasy.input.Input;
import anasy.input.Lexer;
import anasy.input.Location;


/**
 * Base class for representing terms.
 * @author jure
 */
public abstract class Tokind {

	public final String name;
	public final int lbp;
	protected Tokind parent;

	public Tokind(String name, int lbp) {
		this.name = name;
		this.lbp = lbp;
	}

	@Override
	public String toString() {
		return name;
	}

	public void add(Tokind tokind) {
		tokind.setParent(this);
	}

	public void add(Tokind... tokinds) {
		for (Tokind tok : tokinds) add(tok);
	}

	public void setParent(Tokind parent) {
		this.parent = parent;
	}

	/**
	 * Throws a new SyntaxError exception. Given message is processed with String.format(). 
	 * @param loc - location of an error
	 * @param msg - formatted message
	 * @param args - arguments
	 * @throws SyntaxError
	 * @see String.format
	 */
    public void error(Location loc, String msg, Object... args) throws SyntaxError {
    	throw new SyntaxError(String.format(msg, args), loc);
    }

	/**
	 * Return the input used for lexer.
	 * @return input
	 */
	protected Input getInput() {
		return parent.getInput();
	}

	/**
	 * Return the lexer used for parser.
	 * @return lexer
	 */
	protected Lexer getLexer() {
		return parent.getLexer();
	}

	/**
	 * Return recognition precedence that is used when recognizing tokens.
	 * @return recognition precedence
	 */
	public int rp() {
		return 0;
	}

	/** 
	 * Try to recognize a token from the input. If no token (of this tokind) available null is returned.  
	 * @return token
	 * @throws SyntaxError
	 */
    public abstract Token lex() throws SyntaxError;

    /**
     * This method is called after the token is read from the input.
     * @return if token is used, or false if ignored
     */
    public boolean activate() {
    	return true;
    }

    /**
     * Ensured that enough tokens are available. Used in interactive parsing.
     * @throws SyntaxError
     */
	public void ensure() throws SyntaxError {
		parent.ensure();
	}

	protected final boolean isOpening = false;
    protected final boolean isClosing = false;

    /**
     * Called when the token appears at the beginning of an expression.
     * @param rbp - right binding precedence
     * @return token
     * @throws SyntaxError
     */
    public Token parse(int rbp) throws SyntaxError {
	    Token me = getLexer().advance();
		me.error(String.format("unexpected token '%s'", me));
		return me;
	}

    /**
     * Called when the token appears in the middle (or the end) of an expression.
     * @param left - syntax representation of already parsed input
     * @param rbp - right binding precedence
     * @return token
     * @throws SyntaxError
     */
	public Token parse(Token left, int rbp) throws SyntaxError {
	    Token me = getLexer().advance();
		me.error(String.format("unexpected token '%s'", me));
		return me;
	}

	/**
	 * Provides predictable recursive-descent parsing. 
	 * @return token
	 * @throws SyntaxError
	 */
	public Token parse() throws SyntaxError {
    	return parse(0);
    }

}
