package anasy.syntax.literals;

import anasy.Node;
import anasy.SyntaxError;
import anasy.Token;
import anasy.input.Input;
import anasy.input.Location;

public abstract class Number extends Literal {

	public Number(String name) {
		super(name);
	}

	public Number() {
		this("<number>");
	}

	protected abstract Node makeInt(int radix, String lexeme);

	protected abstract Node makeReal(String lexeme);

	@Override
	public Token lex() throws SyntaxError {
		final Input in = getInput();
		if (!Character.isDigit(in.peek())) return null;
		// ***** 0b<bin_digits>, 0o<oct_digits>, 0h<hex_digits>, 0<radix>r<digits>
		final Location l1 = in.getLoc();
		if (in.advanceIf('0')) {	// if leading zero
			int radix = -1;
			switch(in.peek()) {
				case 'b': radix = 2; break;
				case 'o': radix = 8; break;
				case 'h': radix = 16;
				default: if (in.advanceDigits(10) && in.peek() == 'r') {
					int r = Integer.parseInt(in.extract(l1));
					if (r < Character.MIN_RADIX || r > Character.MAX_RADIX || r == 2 || r == 8 || r == 10 || r == 16)
						error(l1, "Invalid radix '%d'", r);
					radix = r;
				}
			}
			if (radix != -1) {	// radix specifier was b,o,h, or r
				in.advance();
				Location l2 = in.getLoc();
				in.advanceDigits(radix);
				// check for proper ending
				if (Character.isLetterOrDigit(in.peek()))
					error(l1, "Invalid digit '%c'", in.peek());
				return new Token(this, l1, makeInt(radix, in.extract(l2)));
			}
		}
		// ***** , <digits>, <digits>.<digits>, <digits>.<digits>[eE][-+]<digits> 
		in.advanceDigits(10);
		// check for fraction part
		boolean isFloat = false;
		if (in.advanceIf('.')) {
			if (!in.advanceDigits(10)) error(in.getLoc(), "Fraction expected");
			isFloat = true;
		}
		// check for exponent part
		if (in.advanceIf('e') || in.advanceIf('E')) {
			if (!in.advanceIf('-')) in.advanceIf('+');
			if (!in.advanceDigits(10)) error(in.getLoc(), "Exponent expected");
			isFloat = true;
		}
		// check for proper ending
		if (Character.isLetterOrDigit(in.peek()))
			error(in.getLoc(), "Invalid digit '%c'", in.peek());
		Node node = isFloat ? makeReal(in.extract(l1)) : makeInt(10, in.extract(l1));
		return new Token(this, l1, node);
	}

}
